"""Domain-level service to delete data related to a dataset."""

from typing import Iterable, Optional

import daiquiri
from dbnomics_data_model import DatasetCode, ProviderCode
from dbnomics_data_model.storage.adapters.filesystem import FileSystemStorage

from dbnomics_solr import services
from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient

__all__ = ["delete_extra_datasets"]


logger = daiquiri.getLogger(__name__)


def delete_extra_datasets(
    provider_code: ProviderCode,
    *,
    dbnomics_solr_client: DBnomicsSolrClient,
    storage: FileSystemStorage,
    storage_dataset_codes: Optional[Iterable[DatasetCode]] = None,
):
    """Delete the extra datasets related to that provider."""
    extra_dataset_codes = services.search_extra_datasets(
        provider_code,
        dbnomics_solr_client=dbnomics_solr_client,
        storage=storage,
        storage_dataset_codes=storage_dataset_codes,
    )
    for dataset_code in extra_dataset_codes:
        dataset_id = f"{provider_code}/{dataset_code}"
        logger.info("Deleting the extra dataset", dataset_id=dataset_id)
        dbnomics_solr_client.delete_dataset_docs(provider_code, dataset_code)
    dbnomics_solr_client.commit()



"""Errors specific to this library."""


from dataclasses import KW_ONLY, dataclass
from typing import Optional

from dbnomics_data_model import DatasetCode, ProviderCode
from solrq import Q

from dbnomics_solr.types import SolrDoc

__all__ = [
    "DatasetAlreadyIndexed",
    "DatasetNotFound",
    "DBnomicsSolrException",
    "DuplicateDocuments",
    "IndexationError",
    "InvalidSolrDocument",
    "ProviderNotFound",
]


class DBnomicsSolrException(Exception):
    pass


@dataclass
class DatasetAlreadyIndexed(DBnomicsSolrException):
    provider_code: ProviderCode
    dataset_code: DatasetCode
    dir_hash: str


@dataclass
class DatasetNotFound(DBnomicsSolrException):
    provider_code: ProviderCode
    dataset_code: DatasetCode


@dataclass
class DuplicateDocuments(DBnomicsSolrException):
    query: Q


@dataclass
class InvalidSolrDocument(DBnomicsSolrException):
    message: str = "Invalid Solr document"
    _: KW_ONLY
    solr_document: SolrDoc

    def __str__(self):
        return self.message


@dataclass
class ProviderNotFound(DBnomicsSolrException):
    provider_slug: Optional[str] = None


@dataclass
class IndexationError(DBnomicsSolrException):
    error: Optional[dict] = None

    def __str__(self):
        return str(self.error)



"""CLI utils."""

__all__ = ["list_callback"]


def list_callback(values: list[str]) -> list[str]:
    """Parse comma-separated values."""

    def iter_values():
        for value in values:
            for part in value.split(","):
                yield part.strip()

    return list(iter_values())



"""Text utils."""

__all__ = ["english_join"]


def english_join(items: list[str]) -> str:
    """Return a string enumerating the items separated by a comma, except for the last one using the "and" separator."""
    return ", ".join(items[:-2] + [" and ".join(items[-2:])])



"""Hash utils."""

from pathlib import Path
from typing import cast

from dirhash import dirhash


def compute_dir_hash(dir_path: Path, /, *, jobs: int = 1) -> str:
    """Compute a hash for a directory in an opinionated way.

    Follows the [dirhash standard](https://github.com/andhus/dirhash).
    """
    return cast(str, dirhash(dir_path, algorithm="sha1", jobs=jobs))



#! /usr/bin/env python3

"""Index DBnomics data into Apache Solr for full-text and faceted search."""


import logging
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import daiquiri
import typer
from dbnomics_data_model import DatasetCode, ProviderCode
from dbnomics_data_model.storage.adapters.filesystem import FileSystemStorage, FileSystemStoragePool
from dotenv import load_dotenv

from . import services
from .cli_utils import list_callback
from .dbnomics_solr_client import DBnomicsSolrClient

INDEXATION = "INDEXATION"
DELETION = "DELETION"
GIT_MODE_TREE_STR = "040000"


@dataclass
class AppArgs:
    """Script arguments common to all commands."""

    solr_timeout: int
    solr_url: str


app = typer.Typer()
app_args: Optional[AppArgs] = None


logger = daiquiri.getLogger(__name__)

load_dotenv()


@app.callback(context_settings={"help_option_names": ["-h", "--help"]})
def main(
    debug: bool = typer.Option(False, help="Display DEBUG log messages."),
    solr_timeout: int = typer.Option(60, envvar="SOLR_TIMEOUT"),
    solr_url: str = typer.Option("http://localhost:8983/solr/dbnomics", envvar="SOLR_URL"),
):
    """Index DBnomics data using Apache Solr."""
    global app_args
    app_args = AppArgs(solr_timeout=solr_timeout, solr_url=solr_url)

    daiquiri.setup()
    daiquiri.set_default_log_levels([(__package__, logging.DEBUG if debug else logging.INFO)])

    logger.debug("Using app args: %r", app_args)


@app.command()
def delete_dataset(
    provider_code: ProviderCode = typer.Option(..., "--provider-code"),
    dataset_code: DatasetCode = typer.Option(..., "--dataset-code"),
):
    """Delete all the documents related to that dataset.

    This includes the document representing the dataset, but also all the ones representing the series of that dataset.
    """
    assert app_args is not None  # it is set by "main" function
    dbnomics_solr_client = DBnomicsSolrClient(app_args.solr_url, timeout=app_args.solr_timeout)
    services.delete_dataset_docs(provider_code, dataset_code, dbnomics_solr_client=dbnomics_solr_client)


@app.command()
def delete_extra_datasets(provider_code: ProviderCode, storage_dir: Path = typer.Option(...)):
    """Delete the extra datasets related to that provider (those that are in Solr but not in the storage)."""
    assert app_args is not None  # it is set by "main" function
    dbnomics_solr_client = DBnomicsSolrClient(app_args.solr_url, timeout=app_args.solr_timeout)
    storage = FileSystemStorage(storage_dir)
    services.delete_extra_datasets(provider_code, dbnomics_solr_client=dbnomics_solr_client, storage=storage)


@app.command()
def delete_provider(
    provider_code: Optional[ProviderCode] = typer.Option(None, "--provider-code"),
    provider_slug: Optional[str] = typer.Option(None, "--provider-slug"),
):
    """Delete all the documents related to that provider.

    This includes the document representing that provider, but also the one representing the datasets of that provider,
    and the ones representing the series of the datasets of that provider.
    """
    assert app_args is not None  # it is set by "main" function

    if bool(provider_code) == bool(provider_slug):
        typer.echo("one of 'code' or 'slug' options must be provided, but not both")
        raise typer.Abort()

    dbnomics_solr_client = DBnomicsSolrClient(app_args.solr_url, timeout=app_args.solr_timeout)

    if provider_code is not None:
        services.delete_provider_docs(provider_code=provider_code, dbnomics_solr_client=dbnomics_solr_client)
    elif provider_slug is not None:
        services.delete_provider_docs(provider_slug=provider_slug, dbnomics_solr_client=dbnomics_solr_client)


@app.command()
def search_datasets(provider_code: ProviderCode):
    """Search the datasets related to that provider."""
    assert app_args is not None  # it is set by "main" function
    dbnomics_solr_client = DBnomicsSolrClient(app_args.solr_url, timeout=app_args.solr_timeout)
    results = services.search_provider_datasets(provider_code, dbnomics_solr_client=dbnomics_solr_client)
    for result in results:
        typer.echo(result["code"])


@app.command()
def search_extra_datasets(provider_code: ProviderCode, storage_dir: Path = typer.Option(...)):
    """Search the extra datasets related to that provider (those that are in Solr but not in the storage)."""
    assert app_args is not None  # it is set by "main" function
    dbnomics_solr_client = DBnomicsSolrClient(app_args.solr_url, timeout=app_args.solr_timeout)
    storage = FileSystemStorage(storage_dir)
    extra_dataset_codes = services.search_extra_datasets(
        provider_code, dbnomics_solr_client=dbnomics_solr_client, storage=storage
    )
    for dataset_code in sorted(extra_dataset_codes):
        typer.echo(dataset_code)


@app.command()
def index_providers(
    storage_base_dir: Path,
    delete_extra_datasets: bool = typer.Option(
        False,
        envvar="DELETE_EXTRA_DATASETS",
        help="After indexation, delete datasets that are in Solr but not in the storage.",
    ),
    delete_obsolete_series: bool = typer.Option(
        False,
        envvar="DELETE_OBSOLETE_SERIES",
        help="After indexation, delete series that were not created or updated.",
    ),
    dirhash_jobs: int = typer.Option(
        1, envvar="DIRHASH_JOBS", help="The number of processes to use when computing a dirhash."
    ),
    fail_fast: bool = typer.Option(False, envvar="FAIL_FAST", help="Disable retry mechanism."),
    force: bool = typer.Option(False, envvar="FORCE", help="Always index data (ignore dir hashes)."),
    limit: Optional[int] = typer.Option(None, envvar="LIMIT", help="Index a maximum number of datasets per provider."),
):
    """Index many providers to Solr from storage_base_dir.

    In storage_base_dir each child directory is considered as the storage directory
    of a provider.
    """
    assert app_args is not None  # it is set by "main" function

    storage_pool = FileSystemStoragePool(storage_base_dir)
    for storage in storage_pool.iter_storages():
        services.index_provider(
            delete_extra_datasets=delete_extra_datasets,
            delete_obsolete_series=delete_obsolete_series,
            dirhash_jobs=dirhash_jobs,
            fail_fast=fail_fast,
            force=force,
            limit=limit,
            solr_timeout=app_args.solr_timeout,
            solr_url=app_args.solr_url,
            storage=storage,
        )


@app.command()
def index_provider(
    storage_dir: Path,
    datasets: List[DatasetCode] = typer.Option(
        [], "--datasets", callback=list_callback, envvar="DATASETS", help="Index only the given datasets."
    ),
    delete_extra_datasets: bool = typer.Option(
        False,
        envvar="DELETE_EXTRA_DATASETS",
        help="After indexation, delete datasets that are in Solr but not in the storage.",
    ),
    delete_obsolete_series: bool = typer.Option(
        False,
        envvar="DELETE_OBSOLETE_SERIES",
        help="After indexation, delete series that were not created or updated.",
    ),
    dirhash_jobs: int = typer.Option(
        1, envvar="DIRHASH_JOBS", help="The number of processes to use when computing a dirhash."
    ),
    excluded_datasets: List[DatasetCode] = typer.Option(
        [],
        "--exclude-datasets",
        callback=list_callback,
        envvar="EXCLUDE_DATASETS",
        help="Do not index the given datasets.",
    ),
    fail_fast: bool = typer.Option(False, envvar="FAIL_FAST", help="Disable retry mechanism."),
    force: bool = typer.Option(False, envvar="FORCE", help="Always index data (ignore dir hashes)."),
    limit: Optional[int] = typer.Option(None, envvar="LIMIT", help="Index a maximum number of datasets."),
    start_from: DatasetCode = typer.Option(None, help="Start indexing from dataset code."),
):
    """Index a single provider to Solr from storage_dir."""
    assert app_args is not None  # it is set by "main" function

    if limit is not None and limit <= 0:
        typer.echo("limit option must be strictly positive")
        raise typer.Abort()

    if not storage_dir.is_dir():
        typer.echo(f"storage_dir {str(storage_dir)} not found")
        raise typer.Abort()

    storage = FileSystemStorage(storage_dir)
    storage_dataset_codes = sorted(storage.iter_dataset_codes(on_error="log"))
    dataset_codes = [
        dataset_code
        for dataset_code in storage_dataset_codes
        if is_desired_dataset(dataset_code, datasets, excluded_datasets, start_from)
    ]

    services.index_provider(
        dataset_codes=dataset_codes,
        delete_extra_datasets=delete_extra_datasets,
        delete_obsolete_series=delete_obsolete_series,
        dirhash_jobs=dirhash_jobs,
        fail_fast=fail_fast,
        force=force,
        limit=limit,
        solr_timeout=app_args.solr_timeout,
        solr_url=app_args.solr_url,
        storage=storage,
        storage_dataset_codes=storage_dataset_codes,
    )


def is_desired_dataset(
    dataset_code: DatasetCode,
    datasets: list[DatasetCode],
    excluded_datasets: list[DatasetCode],
    start_from: DatasetCode,
) -> bool:
    """Apply script arguments to detemine if a dataset has to be indexed."""
    if datasets and dataset_code not in datasets:
        logger.debug("Skipping dataset %r because it is not mentioned by the --datasets option", dataset_code)
        return False
    if excluded_datasets and dataset_code in excluded_datasets:
        logger.debug("Skipping dataset %r because it is mentioned by the --exclude-datasets option", dataset_code)
        return False
    if start_from is not None and dataset_code < start_from:
        logger.debug("Skipping dataset %r because of the --start-from option", dataset_code)
        return False
    return True


if __name__ == "__main__":
    app()



"""Domain-level service to index data related to a provider."""

import time
from datetime import datetime, timezone
from typing import Optional

import daiquiri
from dbnomics_data_model import DatasetCode
from dbnomics_data_model.storage.adapters.filesystem import FileSystemStorage
from dbnomics_data_model.storage.errors import ProviderMetadataNotFound
from humanfriendly import format_timespan
from humanfriendly.text import pluralize

from dbnomics_solr import services
from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient, format_date_for_solr
from dbnomics_solr.errors import DatasetAlreadyIndexed

__all__ = ["index_provider"]


logger = daiquiri.getLogger(__name__)


def index_provider(
    *,
    delete_obsolete_series: bool,
    delete_extra_datasets: bool,
    dirhash_jobs: int,
    fail_fast: bool,
    force: bool,
    solr_url: str,
    dataset_codes: Optional[list[DatasetCode]] = None,
    limit: Optional[int] = None,
    solr_timeout: int = 60,
    storage: FileSystemStorage,
    storage_dataset_codes: Optional[list[DatasetCode]] = None,
):
    """Index data from a provider storage."""
    indexed_at = datetime.now(timezone.utc)
    logger.debug("Using indexed_at %r for all the Solr documents", format_date_for_solr(indexed_at))

    dbnomics_solr_client = DBnomicsSolrClient(
        solr_url,
        dirhash_jobs=dirhash_jobs,
        indexed_at=indexed_at,
        retry_num_attempts=1 if fail_fast else None,
        timeout=solr_timeout,
    )

    # Prepare provider

    try:
        provider_metadata = storage.load_provider_metadata()
    except ProviderMetadataNotFound:
        logger.debug("Provider metadata was not found")
        return

    provider_code = provider_metadata.code
    logger.debug("About to index data for provider %r", provider_code)

    logger.info("Indexing provider metadata...", provider_code=provider_code)
    dbnomics_solr_client.index_provider(provider_metadata)
    dbnomics_solr_client.commit()
    logger.info("Provider metadata indexed", provider_code=provider_code)

    # Index datasets

    if storage_dataset_codes is None:
        storage_dataset_codes = sorted(storage.iter_dataset_codes(on_error="log"))

    if dataset_codes is None:
        dataset_codes = storage_dataset_codes

    if not dataset_codes:
        logger.info("No dataset to process for provider %r", provider_code)
        return

    if limit is not None:
        logger.debug("About to index %d/%d datasets due to the limit option...", limit, len(dataset_codes))
        dataset_codes = dataset_codes[:limit]
    else:
        logger.debug("About to index all the %d datasets...", len(dataset_codes))

    successful_dataset_codes = set()
    skipped_dataset_codes = set()

    for dataset_index, dataset_code in enumerate(dataset_codes, start=1):
        t0 = time.time()
        logger.debug("About to index dataset %r (%d/%d)", dataset_code, dataset_index, len(dataset_codes))
        try:
            dbnomics_solr_client.index_dataset(
                dataset_code, storage=storage, provider_metadata=provider_metadata, force=force
            )
        except DatasetAlreadyIndexed as exc:
            logger.debug(
                "Dataset %r (%d/%d) is already indexed with the directory hash %r "
                "and was not updated since previous indexation, skipping dataset",
                exc.dataset_code,
                dataset_index,
                len(dataset_codes),
                exc.dir_hash,
            )
            continue
        except Exception:
            if fail_fast:
                raise
            logger.exception("Error indexing dataset %r, skipping dataset", dataset_code)
            skipped_dataset_codes.add(dataset_code)
            continue
        else:
            successful_dataset_codes.add(dataset_code)

        dbnomics_solr_client.commit()
        logger.info(
            "Dataset %r (%d/%d) was indexed in %s",
            dataset_code,
            dataset_index,
            len(dataset_codes),
            format_timespan(time.time() - t0),
        )

    if delete_obsolete_series:
        if successful_dataset_codes:
            logger.info(
                "Deleting the obsolete series of the %d successfully indexed datasets...",
                len(successful_dataset_codes),
                provider_code=provider_code,
            )
            dbnomics_solr_client.delete_obsolete_series(provider_code, sorted(successful_dataset_codes))
            dbnomics_solr_client.commit()
            logger.info(
                "The obsolete series of the %d successfully indexed datasets were deleted",
                len(successful_dataset_codes),
                provider_code=provider_code,
            )
        else:
            logger.debug("No dataset was indexed, no need to delete obsolete series")

    if delete_extra_datasets:
        services.delete_extra_datasets(
            provider_code,
            dbnomics_solr_client=dbnomics_solr_client,
            storage=storage,
            storage_dataset_codes=storage_dataset_codes,
        )

    if skipped_dataset_codes:
        logger.error(
            "Summary stats: %s skipped during the indexation process: %r",
            pluralize(len(skipped_dataset_codes), singular="dataset was", plural="datasets were"),
            sorted(skipped_dataset_codes),
            provider_code=provider_code,
        )



from .delete_dataset_docs import *
from .delete_extra_datasets import *
from .delete_provider_docs import *
from .index_provider import *
from .search_extra_datasets import *
from .search_provider_datasets import *



"""Domain-level service to delete data related to a dataset."""

from typing import Iterable, Optional

import daiquiri
from dbnomics_data_model import DatasetCode, ProviderCode
from dbnomics_data_model.storage.adapters.filesystem import FileSystemStorage
from humanfriendly.text import pluralize

from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient

__all__ = ["search_extra_datasets"]


logger = daiquiri.getLogger(__name__)


def search_extra_datasets(
    provider_code: ProviderCode,
    *,
    dbnomics_solr_client: DBnomicsSolrClient,
    storage: FileSystemStorage,
    storage_dataset_codes: Optional[Iterable[DatasetCode]] = None
) -> set[DatasetCode]:
    """Search the extra datasets related to that provider."""
    extra_dataset_codes = dbnomics_solr_client.search_extra_datasets(
        provider_code, storage=storage, storage_dataset_codes=storage_dataset_codes
    )
    if len(extra_dataset_codes) == 0:
        logger.debug("No extra datasets were found for that provider", provider_code=provider_code)
    else:
        logger.debug(
            "Found %s for that provider",
            pluralize(len(extra_dataset_codes), "extra dataset"),
            provider_code=provider_code,
        )
    return extra_dataset_codes



"""Solr client for DBnomics."""

from __future__ import annotations

import logging
import time as t
from dataclasses import dataclass
from datetime import date, datetime, time, timezone
from typing import Any, Iterable, Iterator, Optional, cast
from urllib.parse import quote

import daiquiri
import orjson
import requests
from dbnomics_data_model import ProviderCode
from dbnomics_data_model.model import DatasetCode, DatasetMetadata, ProviderMetadata, SeriesMetadata
from dbnomics_data_model.storage.adapters.filesystem import FileSystemStorage, SeriesJsonLinesOffset
from humanfriendly import format_timespan
from humanfriendly.text import pluralize
from pysolr import Results, Solr
from slugify import slugify
from solrq import Q
from tenacity import Retrying, after_log, retry_if_not_exception_type, stop_after_attempt, wait_random_exponential

from dbnomics_solr.errors import (
    DatasetAlreadyIndexed,
    DatasetNotFound,
    DuplicateDocuments,
    IndexationError,
    InvalidSolrDocument,
    ProviderNotFound,
)
from dbnomics_solr.text_utils import english_join
from dbnomics_solr.types import SolrDoc

from .hash_utils import compute_dir_hash

__all__ = ["DBnomicsSolrClient", "format_date_for_solr"]

logger = daiquiri.getLogger(__name__)


class DBnomicsSolrClient:
    """Solr client for DBnomics."""

    def __init__(
        self,
        solr_url: str,
        *,
        debug_progress_nb_series: int = 10000,
        dirhash_jobs: int = 1,
        indexed_at: Optional[datetime] = None,
        retry_num_attempts: Optional[int] = None,
        timeout: int = 60,
    ):
        """Init DBnomicsSolrClient from Solr URL."""
        self._debug_progress_nb_series = debug_progress_nb_series
        self._dirhash_jobs = dirhash_jobs

        if solr_url.endswith("/"):
            solr_url = solr_url[:-1]
        self._solr = Solr(solr_url, timeout=timeout)
        self._solr_url = solr_url

        if indexed_at is None:
            indexed_at = datetime.now(timezone.utc)
        self._indexed_at = indexed_at

        if retry_num_attempts is None:
            retry_num_attempts = 3
        self._retry_num_attempts = retry_num_attempts

        self._timeout = timeout

    def commit(self):
        """Commit changes to Solr."""
        logger.debug("Committing changes to Solr...")
        self._solr.commit()

    def delete_obsolete_series(self, provider_code: ProviderCode, dataset_codes: Iterable[DatasetCode]):
        """Delete obsolete series related to that dataset.

        Obsolete series are those with a different "indexed_at" property.
        """

        def build_query():
            """Build Solr query matching obsolete documents.

            Hack: build Solr query half with solrq, half manually, because I could not
            achieve working with parentheses generated by solrq when doing
            Q(...) & ~Q(...) => (...) AND (!...)
            Solr did not work with the "!" inside parentheses.
            """
            q1 = Q(type="series", provider_code=provider_code, dataset_code=dataset_code)
            q2 = Q(indexed_at=indexed_at)
            return f"{q1} AND NOT {q2}"

        indexed_at = self._indexed_at

        for dataset_code in dataset_codes:
            query = build_query()
            nb_obsolete_docs = self._solr.search(query).hits
            if nb_obsolete_docs > 0:
                logger.info(
                    "Deleting %d series of dataset %r with indexed_at != %r...",
                    nb_obsolete_docs,
                    dataset_code,
                    format_date_for_solr(indexed_at),
                )
                self._solr.delete(q=query, commit=False)

    def delete_dataset_docs(self, provider_code: ProviderCode, dataset_code: DatasetCode):
        """Delete all the documents related to that dataset.

        This includes the document representing the dataset, but also all the ones representing
        the series of that dataset.
        """
        dataset_id = f"{provider_code}/{dataset_code}"

        logger.info("Deleting the dataset document...", dataset_id=dataset_id)
        self._solr.delete(q=Q(type="dataset", id=dataset_id))

        logger.info("Deleting series documents related to that dataset...", dataset_id=dataset_id)
        self._solr.delete(q=Q(type="series", dataset_id=dataset_id))

    def delete_provider_docs(self, provider_code: ProviderCode):
        """Delete all the documents related to that provider.

        This includes the document representing that provider, but also the one representing the datasets
        of that provider, and the ones representing the series of the datasets of that provider.
        """
        logger.info("Deleting the provider document...", provider_code=provider_code)
        self._solr.delete(q=Q(type="provider", code=provider_code))

        logger.info("Deleting dataset documents related to that provider...", provider_code=provider_code)
        self._solr.delete(q=Q(type="dataset", provider_code=provider_code))

        logger.info("Deleting series documents related to that provider...", provider_code=provider_code)
        self._solr.delete(q=Q(type="series", provider_code=provider_code))

    def search_dataset_docs(self, provider_code: ProviderCode, dataset_code: DatasetCode) -> DBnomicsSolrResults:
        """Search and return the documents related to that dataset.

        This includes the document representing the dataset, but also all the ones representing
        the series of that dataset.
        """
        dataset_id = f"{provider_code}/{dataset_code}"
        dataset_results = self._solr.search(q=Q(type="dataset", id=dataset_id))
        series_results = self._solr.search(q=Q(type="series", dataset_id=dataset_id))
        return DBnomicsSolrResults(dataset_results=dataset_results, series_results=series_results)

    def search_extra_datasets(
        self,
        provider_code: ProviderCode,
        *,
        storage: FileSystemStorage,
        storage_dataset_codes: Optional[Iterable[DatasetCode]],
    ) -> set[DatasetCode]:
        """Search the extra datasets of that provider.

        These are the ones that are in Solr but not in the storage.
        """
        if storage_dataset_codes is None:
            storage_dataset_codes = sorted(storage.iter_dataset_codes(on_error="log"))
        provider_datasets = self.search_provider_datasets(provider_code)
        solr_dataset_codes = {result["code"] for result in provider_datasets}
        return solr_dataset_codes - set(storage_dataset_codes)

    def search_provider_datasets(self, provider_code: ProviderCode) -> Results:
        """Search and return the documents representing the datasets of that provider."""
        return self._solr.search(q=Q(type="dataset", provider_code=provider_code))

    def search_provider_docs(self, provider_code: ProviderCode) -> DBnomicsSolrResults:
        """Search and return the documents related to that provider.

        This includes the document representing that provider, but also the one representing the datasets
        of that provider, and the ones representing the series of the datasets of that provider.
        """
        provider_results = self._solr.search(q=Q(type="provider", code=provider_code))
        dataset_results = self._solr.search(q=Q(type="dataset", provider_code=provider_code))
        series_results = self._solr.search(q=Q(type="series", provider_code=provider_code))
        return DBnomicsSolrResults(
            provider_results=provider_results, dataset_results=dataset_results, series_results=series_results
        )

    def find_provider_code_from_slug(self, provider_slug: str) -> ProviderCode:
        """Return the provider code corresponding to the given slug."""
        provider_solr = self.search_provider_by_slug(provider_slug, strict=False)
        provider_code = provider_solr.get("code")
        if provider_code is None:
            raise InvalidSolrDocument(
                f'Could not find the "code" property in the Solr document of the provider '
                f"which slug is {provider_slug!r}",
                solr_document=provider_solr,
            )
        return provider_code

    def index_dataset(
        self,
        dataset_code: DatasetCode,
        *,
        force: bool = False,
        provider_metadata: ProviderMetadata,
        storage: FileSystemStorage,
    ) -> None:
        """Index a dataset.

        If the dataset found in storage is already indexed in this state, skip it.
        To implement this, a directory hash is computed and stored in Solr.
        """
        provider_code = provider_metadata.code
        storage_dir_hash = self._compute_dir_hash(dataset_code, storage=storage)
        solr_dir_hash = self._fetch_dir_hash_from_solr(provider_code, dataset_code)
        if storage_dir_hash == solr_dir_hash:
            if force:
                logger.debug(
                    "The hash of the directory of the dataset %r (%r) is identical to the one stored in Solr, "
                    'but the dataset will be reindexed due to the "force" option.',
                    dataset_code,
                    storage_dir_hash,
                )
            else:
                raise DatasetAlreadyIndexed(
                    provider_code=provider_code, dataset_code=dataset_code, dir_hash=storage_dir_hash
                )
        else:
            logger.debug(
                "The hash of the directory of the dataset %r (%r) is different from the one stored in Solr (%r), "
                "so the dataset will be reindexed.",
                dataset_code,
                storage_dir_hash,
                solr_dir_hash,
            )

        # Sometimes Solr is too busy and responds an error, so add a retry strategy
        # instead of failing the dataset indexation.
        for attempt in Retrying(
            after=after_log(cast(logging.Logger, logger), logging.ERROR),
            reraise=True,
            retry=retry_if_not_exception_type(DatasetAlreadyIndexed),
            stop=stop_after_attempt(self._retry_num_attempts),
            wait=wait_random_exponential(multiplier=1, max=60 * 5),
        ):
            with attempt:
                json_docs_iter = (
                    serialize_json(without_none_values(solr_doc))
                    for solr_doc in self._iter_docs_for_dataset(
                        dataset_code, storage=storage, provider_metadata=provider_metadata, dir_hash=storage_dir_hash
                    )
                )
                self._index_json_lines(json_docs_iter)

    def index_provider(self, provider_metadata: ProviderMetadata) -> None:
        """Index a provider."""
        provider_code = provider_metadata.code
        provider_solr = without_none_values(
            {
                "id": provider_code,
                "indexed_at": format_date_for_solr(self._indexed_at),
                "type": "provider",
                "code": provider_code,
                "slug": slugify(provider_code),
                "name": provider_metadata.name,
                "region": provider_metadata.region,
                "terms_of_use": provider_metadata.terms_of_use,
                "website": provider_metadata.website,
            }
        )
        self._solr.add(docs=[provider_solr])

    def search_dataset(self, provider_code: ProviderCode, dataset_code: DatasetCode, *, strict: bool = True) -> SolrDoc:
        """Search a dataset document in Solr.

        Raise DatasetNotFound if not found.

        If `strict`, raise a `DuplicateDocuments` error if more than one document is found.
        Otherwise, return the first matching document.
        This allows, for example, the deletion of the documents when there are duplicates.
        It seems that sometimes after an upgrade of Solr, the `uniqueKey` setting of `schema.xml` is ignored
        when adding documents, resulting in many documents having the same slug.
        """
        query = Q(type="dataset", provider_code=provider_code, code=dataset_code)
        result = self._solr.search(query)
        if strict and result.hits > 1:
            raise DuplicateDocuments(query=query)
        if not result:
            raise DatasetNotFound(provider_code=provider_code, dataset_code=dataset_code)
        dataset_solr = result.docs[0]
        return dataset_solr

    def search_provider_by_slug(self, provider_slug: str, *, strict: bool = True) -> SolrDoc:
        """Search a provider document in Solr.

        Raise ProviderNotFound if not found.

        If `strict`, raise a `DuplicateDocuments` error if more than one document is found.
        Otherwise, return the first matching document.
        This allows, for example, the deletion of the documents when there are duplicates.
        It seems that sometimes after an upgrade of Solr, the `uniqueKey` setting of `schema.xml` is ignored
        when adding documents, resulting in many documents having the same slug.
        """
        query = Q(type="provider", slug=provider_slug)
        result = self._solr.search(query)
        if strict and result.hits > 1:
            raise DuplicateDocuments(query=query)
        if not result:
            raise ProviderNotFound(provider_slug=provider_slug)
        dataset_solr = result.docs[0]
        return dataset_solr

    def _compute_dir_hash(self, dataset_code: DatasetCode, *, storage: FileSystemStorage) -> str:
        logger.debug("Computing the hash of the directory of the dataset %r...", dataset_code)
        dataset_dir = storage.get_dataset_dir(dataset_code)
        return compute_dir_hash(dataset_dir, jobs=self._dirhash_jobs)

    def _fetch_dir_hash_from_solr(self, provider_code: ProviderCode, dataset_code: DatasetCode) -> Optional[str]:
        try:
            dataset_solr = self.search_dataset(provider_code, dataset_code)
        except DatasetNotFound:
            return None

        return dataset_solr.get("dir_hash")

    def _iter_series_metadata_and_offsets(
        self, dataset_code: DatasetCode, *, storage: FileSystemStorage
    ) -> Iterator[tuple[SeriesMetadata, Optional[SeriesJsonLinesOffset]]]:
        storage_variant = storage.get_storage_variant(dataset_code)
        if storage_variant == "jsonl":
            dataset_json = storage.load_dataset_json(dataset_code, storage_variant=storage_variant)
            for series, offset in storage.iter_series_jsonl_variant(
                dataset_code,
                dataset_json=dataset_json,
                include_observations=False,
                on_error="log",
            ):
                yield (series.metadata, offset)
            return

        # Offsets are only used with the JSON Lines storage variant.
        # Set them to None otherwise.
        for series in storage.iter_dataset_series(dataset_code, include_observations=False, on_error="log"):
            yield (series.metadata, None)

    def _index_json_lines(self, docs_iter):
        """Index JSON lines from iterator in stream mode.

        pysolr doesn't support this so let's call directly a specific endpoint.
        Cf https://lucene.apache.org/solr/guide/transforming-and-indexing-custom-json.html#TransformingandIndexingCustomJSON-MultipledocumentsinaSinglePayload # noqa
        Cf https://requests.readthedocs.io/en/master/user/advanced/#chunk-encoded-requests
        """
        response = requests.post(
            f"{self._solr_url}/update/json/docs?commit=true", data=docs_iter, timeout=self._timeout
        )
        if not response.ok:
            response_data = response.json()
            error = response_data.get("error")
            raise IndexationError(error=error)

    def _iter_docs_for_dataset(
        self,
        dataset_code: DatasetCode,
        *,
        dir_hash: str,
        provider_metadata: ProviderMetadata,
        storage: FileSystemStorage,
    ):
        """Yield Solr documents for dataset and series for a specific dataset."""
        dataset_id = f"{provider_metadata.code}/{dataset_code}"

        logger.debug("Loading dataset metadata...", dataset_id=dataset_id)
        dataset_metadata = storage.load_dataset_metadata(dataset_code)

        dataset_solr = build_dataset_solr(provider_metadata, dataset_metadata, self._indexed_at, dir_hash)

        logger.debug("Loading series of dataset...", dataset_id=dataset_id)

        t0 = t.time()
        for nb_series_produced, (series_metadata, series_jsonl_offset) in enumerate(
            self._iter_series_metadata_and_offsets(dataset_code, storage=storage), start=1
        ):
            if nb_series_produced % self._debug_progress_nb_series == 0:
                logger.debug("Loaded %d series so far...", nb_series_produced, dataset_id=dataset_id)
            series_solr = build_series_solr(
                provider_metadata,
                dataset_metadata,
                series_metadata,
                indexed_at=self._indexed_at,
                series_jsonl_offset=series_jsonl_offset,
            )
            yield series_solr
            dataset_solr["nb_series"] += 1

        logger.debug(
            "Loaded a total of %d series in %s",
            dataset_solr["nb_series"],
            format_timespan(t.time() - t0),
            dataset_id=dataset_id,
        )

        # Obsolete series of the current dataset will be removed later.

        yield dataset_solr


@dataclass
class DBnomicsSolrResults:
    dataset_results: Results
    series_results: Results
    provider_results: Optional[Results] = None

    def format_hits_by_type(self) -> str:
        items = []
        if self.provider_results is not None:
            items.append(pluralize(self.provider_results.hits, "provider document"))
        items.extend(
            [
                pluralize(self.dataset_results.hits, "dataset document"),
                pluralize(self.series_results.hits, "series document"),
            ]
        )
        return english_join(items)

    def hit_sum(self) -> int:
        provider_hits = self.provider_results.hits if self.provider_results is not None else 0
        return provider_hits + self.dataset_results.hits + self.series_results.hits


def build_dataset_solr(
    provider_metadata: ProviderMetadata, dataset_metadata: DatasetMetadata, indexed_at, dir_hash: str
) -> SolrDoc:
    """Build a JSON object for a dataset, following the Solr schema."""
    provider_code = provider_metadata.code
    dataset_code = dataset_metadata.code

    updated_at = dataset_metadata.updated_at
    updated_at_str = None if updated_at is None else format_date_for_solr(updated_at)

    dataset_solr = {
        "id": f"{provider_code}/{dataset_code}",
        "indexed_at": format_date_for_solr(indexed_at),
        "type": "dataset",
        "code": dataset_code,
        "provider_code": provider_code,
        "provider_name": provider_metadata.name,
        "name": dataset_metadata.name,
        "description": dataset_metadata.description,
        "dir_hash": dir_hash,
        "updated_at": updated_at_str,
        "nb_series": 0,  # Will be incremented if dataset has series.
    }

    return dataset_solr


def build_series_solr(
    provider_metadata: ProviderMetadata,
    dataset_metadata: DatasetMetadata,
    series_metadata: SeriesMetadata,
    *,
    indexed_at: datetime,
    series_jsonl_offset: Optional[SeriesJsonLinesOffset],
) -> SolrDoc:
    """Build a JSON object for a series, following the Solr schema."""
    provider_code = provider_metadata.code
    dataset_code = dataset_metadata.code
    series_code = series_metadata.code
    series_solr: SolrDoc = {
        "id": f"{provider_code}/{dataset_code}/{series_code}",
        "indexed_at": format_date_for_solr(indexed_at),
        "type": "series",
        "code": series_code,
        "name": series_metadata.name,
        "provider_code": provider_code,
        "provider_name": provider_metadata.name,
        "dataset_id": f"{provider_code}/{dataset_code}",
        "dataset_code": dataset_code,
        "dataset_name": dataset_metadata.name,
    }
    if series_jsonl_offset is not None:
        series_solr["series_jsonl_offset"] = series_jsonl_offset

    for dimension_code, dimension_value_code in series_metadata.dimensions.items():
        # Index dimension codes to compute facets.
        series_solr[f"dimension_value_code.{quote(dimension_code)}"] = dimension_value_code

    dimensions_values_labels = list(iter_dimensions_values_labels(series_metadata, dataset_metadata=dataset_metadata))
    if dimensions_values_labels:
        series_solr["dimensions_values_labels"] = dimensions_values_labels

    return series_solr


def iter_dimensions_values_labels(
    series_metadata: SeriesMetadata, *, dataset_metadata: DatasetMetadata
) -> Iterator[str]:
    for dimension_code, dimension_value_code in series_metadata.dimensions.items():
        dimension_def = dataset_metadata.find_dimension_by_code(dimension_code)
        if dimension_def is None:
            continue
        dimension_value_def = dimension_def.find_value_by_code(dimension_value_code)
        if dimension_value_def is None:
            continue
        dimension_value_label = dimension_value_def.label
        if dimension_value_label is not None:
            yield dimension_value_label


def format_date_for_solr(d: date) -> str:
    """Format a `date` or a `datetime` for Solr.

    If `d` is a `date` (and not a `datetime`), assume the time is `00:00:00` in the UTC timezone.

    Solr supports a subset of ISO-8601 and requires the datetime to be expressed in the UTC timezone
    with the "Z" suffix notation instead of the "+00:00" generated by Python.

    If the timezone of a datetime is not specified, UTC is assumed and a warning is logged.

    >>> from datetime import timedelta
    >>> format_date_for_solr(date(2000, 1, 1))
    '2000-01-01T00:00:00Z'
    >>> format_date_for_solr(datetime(2000, 1, 1, 5, 5, 5))
    '2000-01-01T05:05:05Z'
    >>> format_date_for_solr(datetime(2000, 1, 1, 5, 5, 5, tzinfo=timezone.utc))
    '2000-01-01T05:05:05Z'
    >>> format_date_for_solr(datetime(2000, 1, 1, 5, 5, 5, tzinfo=timezone(timedelta(hours=2))))
    '2000-01-01T03:05:05Z'
    """
    if not isinstance(d, datetime):
        return datetime.combine(d, time.min).isoformat() + "Z"

    if d.tzinfo is None:
        logger.warning("No timezone defined in %r, assuming it is UTC", d)
        d = d.replace(tzinfo=timezone.utc)
    else:
        d = d.astimezone(timezone.utc)

    return d.isoformat().removesuffix("+00:00") + "Z"


def serialize_json(data: Any) -> bytes:
    return orjson.dumps(data)


def without_none_values(d):
    """Return a copy of d without None values."""
    return {k: v for k, v in d.items() if v is not None}



"""Domain-level service to delete data related to a dataset."""

import daiquiri
from dbnomics_data_model import DatasetCode, ProviderCode

from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient

__all__ = ["delete_dataset_docs"]


logger = daiquiri.getLogger(__name__)


def delete_dataset_docs(
    provider_code: ProviderCode, dataset_code: DatasetCode, *, dbnomics_solr_client: DBnomicsSolrClient
):
    """Delete Solr documents related to that dataset."""
    dataset_id = f"{provider_code}/{dataset_code}"

    results = dbnomics_solr_client.search_dataset_docs(provider_code, dataset_code)
    if results.hit_sum() == 0:
        logger.warning(
            "No documents related to that dataset were found, but sending the delete requests anyway.",
            dataset_id=dataset_id,
        )
    else:
        logger.debug("Found %s", results.format_hits_by_type())

    logger.info("Deleting all the Solr documents related to that dataset...", dataset_id=dataset_id)
    dbnomics_solr_client.delete_dataset_docs(provider_code, dataset_code)

    dbnomics_solr_client.commit()
    logger.info("All the Solr documents related to that dataset were deleted", dataset_id=dataset_id)



"""Domain-level service to delete data related to a provider."""

from typing import Optional

import daiquiri
from dbnomics_data_model import ProviderCode

from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient
from dbnomics_solr.errors import InvalidSolrDocument, ProviderNotFound

__all__ = ["delete_provider_docs"]


logger = daiquiri.getLogger(__name__)


def delete_provider_docs(
    *,
    provider_code: Optional[ProviderCode] = None,
    provider_slug: Optional[str] = None,
    dbnomics_solr_client: DBnomicsSolrClient
):
    """Delete Solr documents related to a provider identified by its code or its slug."""
    if provider_code is None and provider_slug is None:
        raise ValueError("Either provider_code or provider_slug must be specified")

    if provider_slug is not None:
        logger.debug("Finding the provider code from the provider slug...", provider_slug=provider_slug)
        assert provider_code is None, provider_code
        try:
            provider_code = dbnomics_solr_client.find_provider_code_from_slug(provider_slug)
        except InvalidSolrDocument:
            logger.error(
                "Could not find the provider code in the Solr document for the provider, skipping deletion",
                provider_slug=provider_slug,
            )
            return
        except ProviderNotFound:
            logger.error(
                "Could not find the Solr document for the provider, skipping deletion", provider_slug=provider_slug
            )
            return

    assert provider_code is not None

    results = dbnomics_solr_client.search_provider_docs(provider_code)
    if results.hit_sum() == 0:
        logger.warning(
            "No documents related to that provider were found, but sending the delete requests anyway.",
            provider_code=provider_code,
        )
    else:
        logger.debug("Found %s", results.format_hits_by_type())

    logger.info("Deleting all the Solr documents related to that provider...", provider_code=provider_code)
    dbnomics_solr_client.delete_provider_docs(provider_code)

    dbnomics_solr_client.commit()
    logger.info("All the Solr documents related to that provider were deleted", provider_code=provider_code)



"""Domain-level service to delete data related to a dataset."""

import daiquiri
from dbnomics_data_model import ProviderCode
from humanfriendly.text import pluralize
from pysolr import Results

from dbnomics_solr.dbnomics_solr_client import DBnomicsSolrClient

__all__ = ["search_provider_datasets"]


logger = daiquiri.getLogger(__name__)


def search_provider_datasets(provider_code: ProviderCode, *, dbnomics_solr_client: DBnomicsSolrClient) -> Results:
    """Search the dataset documents related to that provider."""
    results = dbnomics_solr_client.search_provider_datasets(provider_code)
    if results.hits == 0:
        logger.debug("No dataset documents were found for that provider", provider_code=provider_code)
    else:
        logger.debug(
            "Found %s for that provider", pluralize(results.hits, "dataset document"), provider_code=provider_code
        )
    return results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbnomics_solr', 'dbnomics_solr.services']

package_data = \
{'': ['*']}

install_requires = \
['daiquiri>=3.0.1,<4.0.0',
 'dbnomics-data-model>=0.13.31,<0.14.0',
 'dirhash>=0.2.1,<0.3.0',
 'humanfriendly>=10.0,<11.0',
 'orjson>=3.6.7,<4.0.0',
 'pysolr>=3.9.0,<4.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'python-slugify>=6.1.1,<7.0.0',
 'requests>=2.27.1,<3.0.0',
 'solrq>=1.1.1,<2.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['dbnomics-solr = dbnomics_solr.cli:app']}

setup_kwargs = {
    'name': 'dbnomics-solr',
    'version': '1.1.14',
    'description': 'Index DBnomics data with Apache Solr for full-text and faceted search',
    'long_description': '# DBnomics Solr\n\nIndex DBnomics data into Apache Solr for full-text and faceted search.\n\nRequirements:\n\n- a running instance of [Apache Solr](http://lucene.apache.org/solr/); at the time this documentation is written, we use the version 7.3.\n\nSee [dbnomics-docker](https://git.nomics.world/dbnomics/dbnomics-docker) to run a local DBnomics instance with Docker that includes a service for Apache Solr.\n\n## Configuration\n\nEnvironment variables:\n\n- `DEBUG_PYSOLR`: display pysolr DEBUG logging messages (cf <https://github.com/django-haystack/pysolr>)\n\n## Index a provider\n\nReplace `wto` by the real provider slug in the following command:\n\n```bash\ndbnomics-solr index-provider /path/to/wto-json-data\n```\n\n### Full mode vs incremental mode\n\nWhen data is stored in a regular directory, the script always indexes all datasets and series of a provider. This is called _full mode_.\n\nWhen data is stored in a Git repository, the script runs by default in _incremental mode_: it indexes only the datasets modified since the last indexation.\n\nIt is possible to force the _full mode_ with the `--full` option.\n\n### Bare repositories\n\nThe script has an option `--bare-repo-fallback` which tries to add `.git` at the end of the storage dir name, if not found.\n\n## Remove all data from a provider\n\nTo remove all the documents related to a provider (`type:provider`, `type:dataset` and `type:series`):\n\n```bash\ndbnomics-solr --debug delete-provider --code <provider_code>\ndbnomics-solr --debug delete-provider --slug <provider_slug>\n\n# Examples:\ndbnomics-solr --debug delete-provider --code WTO\ndbnomics-solr --debug delete-provider --slug wto\n```\n',
    'author': 'Christophe Benz',
    'author_email': 'christophe.benz@nomics.world',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://git.nomics.world/dbnomics/dbnomics-solr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
